@@ -7,6 +7,8 @@ import inspect
 from pathlib import Path
 from typing import Any, Awaitable, Callable, Optional, ParamSpec, TypeVar, overload
 
+from pydantic import TypeAdapter, ValidationError
+
 from .compression import CompressionPolicy
 from .llm import LLMConfig, call_llm
 from .memory import MemoryManager
@@ -20,6 +22,35 @@ P = ParamSpec("P")
 T = TypeVar("T")
 
 
+def _validate_return_value(result: Any, return_type: Any, tool_name: str) -> Any:
+    """Validate tool return value against annotated type.
+
+    Args:
+        result: The returned value from the tool
+        return_type: The annotated return type
+        tool_name: Name of the tool (for error messages)
+
+    Returns:
+        The validated result
+
+    Raises:
+        TypeError: If validation fails
+    """
+    if return_type == inspect.Signature.empty:
+        return result
+
+    try:
+        # Use Pydantic TypeAdapter for validation
+        adapter: TypeAdapter[Any] = TypeAdapter(return_type)
+        validated = adapter.validate_python(result)
+        return validated
+    except ValidationError as e:
+        raise TypeError(
+            f"Tool '{tool_name}' returned invalid type. "
+            f"Expected {return_type}, but validation failed: {e}"
+        ) from e
+
+
 def _convert_tools_to_llm_format(tools: list[Callable]) -> list[dict[str, Any]]:
     """Convert Python tool functions to LiteLLM tools format.
 
@@ -640,18 +671,16 @@ async def _execute_agent(
 
     # Record LLM call in telemetry
     if telemetry_collector:
-        # Import LLMResponse for type checking
-        from kagura.core.llm import LLMResponse
-
-        if isinstance(response, LLMResponse):
+        # Check if response is LLMResponse (has content, usage, model, duration)
+        if hasattr(response, "content") and hasattr(response, "usage"):
             from kagura.observability.pricing import calculate_cost
 
             telemetry_collector.record_llm_call(
-                model=response.model,
-                prompt_tokens=response.usage.get("prompt_tokens", 0),
-                completion_tokens=response.usage.get("completion_tokens", 0),
-                duration=response.duration,
-                cost=calculate_cost(response.usage, response.model),
+                model=response.model,  # type: ignore
+                prompt_tokens=response.usage.get("prompt_tokens", 0),  # type: ignore
+                completion_tokens=response.usage.get("completion_tokens", 0),  # type: ignore
+                duration=response.duration,  # type: ignore
+                cost=calculate_cost(response.usage, response.model),  # type: ignore
             )
 
     # Parse response based on return type annotation
@@ -661,6 +690,10 @@ async def _execute_agent(
         response_str = str(response)
         return parse_response(response_str, return_type)  # type: ignore
 
+    # If return type is str, extract content from LLMResponse
+    if hasattr(response, "content"):
+        return response.content  # type: ignore
+
     return response  # type: ignore
 
 
@@ -827,9 +860,7 @@ def tool(
 
                 # Validate return type if annotated
                 return_type = sig.return_annotation
-                if return_type != inspect.Signature.empty:
-                    # TODO: Add Pydantic validation for return type
-                    pass
+                result = _validate_return_value(result, return_type, tool_name)
 
                 return result  # type: ignore
 
@@ -852,9 +883,7 @@ def tool(
 
                 # Validate return type if annotated
                 return_type = sig.return_annotation
-                if return_type != inspect.Signature.empty:
-                    # TODO: Add Pydantic validation for return type
-                    pass
+                result = _validate_return_value(result, return_type, tool_name)
 
                 return result  # type: ignore
 
